/*
 * Copyright (c) 2008 Jing Su. All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
 */
/*
 credits: Lance Fetters
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <signal.h>
#include <stdlib.h>

typedef uint32_t uint32;		// this annoys me too
#include "iphone.h"

#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <resolv.h>
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>



#define BACKLOG 10
#define DEFAULT_TIMEOUT 4

typedef struct {
    iphone_device_t phone;

} itunnel_t;

typedef struct {
    volatile int dead;
    pthread_t thread;
    pthread_t reader;
    int reader_quit; // hack flag to signal quit to the reader

    int fd;
    struct sockaddr_in peeraddr;
    socklen_t addrlen;
    iphone_umux_client_t usb;

} client_params_t;

static char* g_serviceport = "3023";
static int g_noreset = 0;
static itunnel_t g_context;
static int quit_flag = 0;

/*********************************************************************
 */
static void release_iphone() {

    if (g_context.phone != NULL) {
        iphone_free_device(g_context.phone);
        g_context.phone = NULL;
    }
}

/*********************************************************************
 * @return 1 for success
 * @return 0 for error
 */
static int init_iphone() {
    g_context.phone = NULL;

    iphone_error_t err = iphone_get_device(&g_context.phone);
    if (err != IPHONE_E_SUCCESS) {
        fprintf(stderr, "- iphone get device error: %hd\n", err);
        return 0;
    }
    if (g_context.phone == NULL)
        return 0;
    fprintf(stdout, "- successfully got device\n");

    return 1;
}


/**********************************************************************/
static void* clientthread_reader(void* arg) {
    // pull parameters
    client_params_t* param = (client_params_t*)arg;
    
    fd_set wfds;
    struct timeval tv;

    char rbuffer[512];
    uint32_t rbuffersize = 512;

    while(!quit_flag && !param->reader_quit) {

        int maxfd = param->fd;

        FD_ZERO(&wfds);
        FD_SET(param->fd, &wfds);
        
        tv.tv_sec = DEFAULT_TIMEOUT;
        tv.tv_usec = 0;

        int result = select(maxfd+1, NULL, &wfds, NULL, &tv);
        if (result == -1) {
            fprintf(stderr, "client thread fd(%d) encountered select() error:%d\n", param->fd, errno);
            continue;
        }

        if (result > 0 && FD_ISSET(param->fd, &wfds)) {

            uint32_t rlen = 0;
            iphone_error_t err = IPHONE_E_SUCCESS;

            err = iphone_mux_recv_timeout(param->usb, rbuffer, rbuffersize, &rlen, DEFAULT_TIMEOUT);
            if (err != IPHONE_E_SUCCESS) {
                fprintf(stdout, "client thread fd(%d) encountered USB read error: %hd\n", param->fd, err);
                break;
            }
            char* cursor = rbuffer;
            while(rlen > 0 && !quit_flag) {
                ssize_t len = write(param->fd, cursor, rlen);
                // calculate remainder
                rlen = rlen - len;
                // advance cursor
                cursor = &cursor[len];
            }
            fsync(param->fd);
        }
    }
    fprintf(stdout, "client thread fd(%d) reader ended.\n", param->fd);
    return NULL;
}

/**********************************************************************/
static void* clientthread(void* arg) {
    // pull parameters
    client_params_t* param = (client_params_t*)arg;
    param->dead = 0;

    // get name info (local server side) of the socket
    struct sockaddr_in localaddr = {0};
    socklen_t localaddrlen;
    localaddrlen = sizeof(localaddr);
    if (getsockname(param->fd, (struct sockaddr*)&localaddr, &localaddrlen) == -1) {
        fprintf(stderr, "client thread fd(%d) unable to get local address. errno:%d\n", param->fd, errno);
    }
    fprintf(stdout, "client thread fd(%d) running, server port %d , peer port %d\n", 
            param->fd, ntohs(localaddr.sin_port), ntohs(param->peeraddr.sin_port));

    // ugh... I hate this.
    // I just can't bring myself to use a damn GOTO, so as a result
    // I have an extra sentry flag here which terminates the loop.
    // it's mostly used right at the start, in case USBMUX fails.
    int sentry = 1;

    // establish USB MUX connection with iphone
    iphone_error_t err = IPHONE_E_SUCCESS;
    param->usb = NULL;
    uint16_t srcport = ntohs(param->peeraddr.sin_port);
    const uint16_t destport = 22;
    err = iphone_mux_new_client(g_context.phone, srcport, destport, &param->usb);    
    if (IPHONE_E_SUCCESS != err || param->usb == NULL) {
        fprintf(stderr, "client thread fd(%d) failed to get USBMUX client. err=%hd\n", param->fd, err);
        sentry = 0;
    }

    // usb reader thread
    param->reader_quit = 0;
    param->reader = -1;
    int result = pthread_create(&param->reader, NULL, clientthread_reader, param);
    if (result == 0) {
        // ok
    }
    else {
        fprintf(stderr, "client thread fd(%d) failed to spawn USB reader.\n", param->fd);
        sentry = 0;
        param->reader = -1;
    }
    

    // all is good.  some last minute vars before we continue
    // the buffer
    char buffer[1024];
    ssize_t maxlen = 1024;
    // fd select vars
    fd_set rfds;
    struct timeval tv;
    // ok.
    while(!quit_flag && sentry) {

        int maxfd = param->fd;

        FD_ZERO(&rfds);
        FD_SET(param->fd, &rfds);
        tv.tv_sec = DEFAULT_TIMEOUT;
        tv.tv_usec = 0;

        int result = select(maxfd+1, &rfds, NULL, NULL, &tv);
        if (result == -1) {
            fprintf(stderr, "client thread fd(%d) encountered select() error:%d\n", param->fd, errno);
            continue;
        }

        if (result > 0 && FD_ISSET(param->fd, &rfds)) {
            // Read from client connection.  Push to USB

            ssize_t len = read(param->fd, buffer, maxlen);
            if (len == 0) break; // EOF
            if (len < 0) {
                fprintf(stdout, "client thread fd(%d) encountered read error: %d\n", param->fd, errno);
                break;
            }

            char* cursor = buffer;
            do {
                uint32_t wlen = 0;
                err = iphone_mux_send(param->usb, cursor, len, &wlen);
                if (err == IPHONE_E_TIMEOUT) {
                    // some kind of timeout.... just be patient and retry.
                }
                else if (err != IPHONE_E_SUCCESS) {
                    fprintf(stdout, "client thread fd(%d) encountered USB write error: %hd\n", param->fd, err);
                    len = -1; // signal error
                    break;
                }

                // calculate remainder.
                len -= wlen;
                // advance cursor appropriately.
                cursor += wlen;
            } while(len > 0 && !quit_flag);
            if (len < 0) break;
        }        
    }

    // cleanup
    fprintf(stdout, "client thread fd(%d) terminating\n", param->fd);
    if (param->reader != -1) {
        param->reader_quit = 1;
        pthread_join(param->reader, NULL);
        param->reader = -1;
    }
    
    close(param->fd);
    if (param->usb != NULL) {
        iphone_mux_free_client(param->usb);
    }

    param->dead = 1;
    return NULL;
}


/**********************************************************************/
static void server_loop(int socketfd) {

    int reinit = 1;

    while(!quit_flag) {

        if (reinit) {
            release_iphone();

            if (!init_iphone()) {
                fprintf(stderr, "ERROR: failed to connect to iphone. quitting.\n");
                quit_flag = 1;
                break;
            }
        }
        if (g_noreset) reinit = 0;

        // if we get here, it means a client tried to connect. accept.
        // since we want to pass a bunch of this connection info to the
        // client, go a head and make the client param struct here.
        client_params_t* param = (client_params_t*)malloc(sizeof(client_params_t));
        memset(param, 0, sizeof(client_params_t));
        param->addrlen = sizeof(param->peeraddr);

        fprintf(stdout, "server waiting for ssh connection\n");
        int clientfd = accept(socketfd, (struct sockaddr*)&param->peeraddr, &param->addrlen);
        if (clientfd < 0) {
            // we have an exception here.
            // if the user is killing the server, the we get the Interrupted System Call
            // error return value.
            if (errno == EINTR) {
                // ignore... don't scare the user.  This is okay.
            }
            else {
                fprintf(stderr, "accept returned error: %d\n", errno);
            }
            free(param); // make sure we clean up.
            continue;
        }
        
        fprintf(stdout, "server accepted connection, clientfd:%d\n", clientfd);
        param->fd = clientfd;            
        int result = pthread_create(&param->thread, NULL, clientthread, param);
        if (result != 0) {
            fprintf(stderr, "failed to create thread. cleaning up.\n");
            free(param);
            param = NULL;
            close(clientfd);                
            continue;
        }

        fprintf(stdout, "tunnel now running.\n");
        fprintf(stdout, "SSH to localhost port %s to get to iPhone.\n", g_serviceport);
        fprintf(stdout, " (use ssh -D <port> to create SOCKS tunnel.)\n");
        fprintf(stdout, "do Ctrl+C to quit the server (can take a few seconds).\n");

        // at this point we need to go into a tight loop that pumps USB bulk
        // packets.
        while (!param->dead && !quit_flag) {
            iphone_mux_pullbulk(g_context.phone);
        }

        // okay once we drop out here, the client is closed.
        // make sure we wait on our child.
        fprintf(stdout, "server waiting for client thread join...");
        pthread_join(param->thread, NULL);
        free(param);
        fprintf(stdout, " ok.\n");
    }

    // call the release on the iphone to make sure it's closed up and clean
    // before we exit this function.
    release_iphone();
}

/*********************************************************************
 * @return 1 for success
 * @return 0 for failure
 */
int start_server(void) {

    int result = 0;

    // before we can do the socket, much song and dance... :(
    struct addrinfo hints;
    bzero(&hints, sizeof(struct addrinfo));
    hints.ai_family=AF_INET; // ipv4
    hints.ai_flags=AI_PASSIVE;
    hints.ai_socktype=SOCK_STREAM;
    hints.ai_protocol=IPPROTO_TCP;
    //
    struct addrinfo* res = NULL;
    int num_interfaces = getaddrinfo(NULL, g_serviceport, &hints, &res);
    if (num_interfaces != 0) {
        fprintf(stderr, "failed to getaddrinfo. error code=%d\n", num_interfaces);
        return 0;
    }

    // unfortunately, socket creation is a little bit tricky.
    // we need to loop through the available interfaces to find one
    // that we can bind to.
    int socketfd = -1;
    struct addrinfo* cursor = res;
    for (; cursor != NULL; cursor = cursor->ai_next) {
        // create the socket here.
        socketfd = socket(cursor->ai_family, cursor->ai_socktype, cursor->ai_protocol);
        if (socketfd == -1) {
            continue; // nope... skip.
        }
        int on = 1; // literally, a value to indicate enable
        setsockopt(socketfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

        // attempt the bind
        result = bind(socketfd, cursor->ai_addr, cursor->ai_addrlen);
        if (result == 0) {
            // success
            break;
        }

        // close and continue
        close(socketfd);
        socketfd = -1;
    }

    // free up the dynamically allocated res list.
    freeaddrinfo(res);
    res = NULL;

    // socket result check
    if (socketfd == -1) {
        fprintf(stderr, "failed to find bindable address for server socket.\n");
        return 0;
    }

    // listen for connections
    result = listen(socketfd, BACKLOG);
    if (result < 0) {
        fprintf(stderr, "failed to initiate listening. errno=%d\n", errno);
        close(socketfd);
        return 0;
    }

    // run the server loop
    server_loop(socketfd);
    close(socketfd);
    return 1;
}

/**********************************************************************/
void signal_handler(int foo) {
    quit_flag = 1;
    fprintf(stdout, "sent out quit signal... waiting for cleanups and joins.\n");
}

/**********************************************************************/
int main(int argc, char* argv[]) {

    // command line argument processing
    // TODO: I need to figure out how to use libraries to help with this.
    int argi = 1;
    for (; argi < argc; argi++) {
        char* thisarg = argv[argi];
        if (thisarg[0] == '\0') continue; // paranoia

        if (thisarg[0] == '-') {
            if (strcmp(thisarg, "-h")==0 || strcmp(thisarg, "-help")==0) {
                fprintf(stdout, "usage: %s [opts] [listenport]\n\n", argv[0]);
                fprintf(stdout, " options:\n");
                fprintf(stdout, "   -noreset : do not reset the USB after each connection.\n");
                fprintf(stdout, "\n");
                return 0;
            }
            else if (strcmp(thisarg, "-noreset") == 0) {
                g_noreset = 1;
            }
            else {
                fprintf(stderr, "unrecognized argument: %s\n\n", thisarg);
                return 1;
            }
        }
        else {
            g_serviceport = thisarg;

            // the service port parameter is always the last parameter.
            // we ignore all other params that come after this.
            break;
        }
    }

    // set up signal handler
    struct sigaction sa;
    sa.sa_handler = signal_handler;
    sa.sa_flags = 0;
    sigemptyset(&sa.sa_mask);
    if (sigaction(SIGINT, &sa, NULL) < 0 ||
            sigaction(SIGHUP, &sa, NULL) < 0 ||
            sigaction(SIGTERM, &sa, NULL) < 0) {
        fprintf(stderr, "ERROR: Unable to setup signal handler.\n");
        return 1;
    }

    // run the single-connection server and its loop
    if (start_server() == 0) {
        fprintf(stderr, "ERROR: failed to start the server socket.\n");
        return 1;
    }

    fprintf(stdout, "quit ok.\n");
    return 0;
}
